#!/usr/bin/perl -w

#Murasaki - multiple genome global alignment program
#Copyright (C) 2006-2007 Keio University
#(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)
#
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

use File::Basename;
use Bio::SeqIO;
use Getopt::Long;
use Pod::Usage;
#use Data::Dump qw{dump};

use strict;
my ($help,$man,$opt_prefix);

our ($seqhome,$root,$flexible);
$root=(fileparse($0))[1];
$root=$ENV{MURASAKI} if $ENV{MURASAKI};
require "${root}/config.pl";
require "$root/common.pl";

our %common_overrides;
our %redirect;

my $opt_force;
GetOptions('help|?' => \$help, man => \$man, flexible => \$flexible, force=>\$opt_force);
pod2usage(1) if $help or $#ARGV<0;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

my $startdir=$ENV{PWD};

while(@ARGV){
  my ($filename,$outfile);
  $filename=shift(@ARGV);
  $outfile=shift(@ARGV) if @ARGV;
  if($outfile and -f $outfile){ #if it's another input file, throw it back on!
    unshift(@ARGV,$outfile);
    $outfile="$filename.cds";
  }

  $outfile="$filename.cds" unless $outfile;
  my @cds_features;
  our $src_primary='CDS';
  open(my $infh,$filename) or die "Couldn't open source file $filename";
  procOverrides("$filename.overrides") if -f "$filename.overrides";
  $src_primary=$common_overrides{CDS} if $common_overrides{CDS};

  my $redirect="$filename.redirect";
  my $globalRedirect=getPath($redirect)."/.redirect";
  push(@cds_features,useRedirect($filename,$globalRedirect,$outfile)) if -f $globalRedirect;
  push(@cds_features,useRedirect($filename,$redirect,$outfile)) if -f $redirect;

  $_=<$infh>;
  if(m/(\S+)\t(\d+)\t(\d+)\t(\d+)/){ #stitch format
    print "$filename is a stitch file. Compiling constituent parts:\n";
    do{
      my ($src,$len,$start,$stop)=m/(\S+)\t(\d+)\t(\d+)\t(\d+)/;
      #recursively generate baby!
      my $localpath=getPath($filename);
      if(!-f $src and $localpath ne './'){
	print "Don't see $src from this directory.\n".
	  "Trying from $localpath\n";
	chdir($localpath);
      }
      die "Can't find component file: $src\n" unless -f $src;
      my $runopts="";
      $runopts.=" --flexible" if $flexible;
      $runopts.=" --force" if $opt_force;
      system("$root/getcds.pl $runopts $src") unless -f "$src.cds";
      die "Couldn't load annotation data for $src" unless -f "$src.cds";
      my $offset=$start-1;
      print "Adding $src.cds with offset $offset -> ";
      my @cds=readCds("$src.cds",$offset);
      print scalar(@cds)." CDS's\n";
      push(@cds_features,@cds);
    }while(<$infh>);
    close($infh);
  }else{ #simple file
    unless($redirect{nobioperl}){
      close($infh);
      print "Loading $filename\n";
      my $seq=Bio::SeqIO->new(-file => $filename)->next_seq;
      die "Error loading $filename"  unless $seq;
      @cds_features = grep { $_->primary_tag eq $src_primary } $seq->get_SeqFeatures;
      print "$filename loaded. Contains ".($#cds_features+1)." CDSs\n";
      if($#cds_features<0 and !$opt_force){
	print "Cowardly refusing to write an empty .cds file (use --force to force this).\n";
	exit(2);
      }
    }
  }
  writeCds("$outfile",@cds_features);
}

sub useRedirect {
  my ($filename,$redirect,$outfile)=@_;
  print "Loading redirect $redirect...\n";
  our $src_primary;
  open(my $infh,$redirect) or return;
  while(<$infh>){
    chomp;
    s!^(.*?)(#|//|;|%).*$!$1!; #chop off various comments
    my ($key,$val)=m/^(.*?)=(.*)$/;
    next unless $key; #nothing? skip...
    $val=~s/\s+$//; #cleanup any spaces at end
    print "Redirect: $key : $val\n";
    $redirect{$key}=$val;
  }
  my ($chromosome,$offset)=(0,0);
  my $raw_chromosome; #before getting mangled hacchy-style
  my $out_primary=$src_primary;
  $out_primary=$redirect{primary} if exists $redirect{primary};
  $offset=$redirect{offset} if exists $redirect{offset};
  $redirect{nobioperl}=1 unless exists $redirect{'+bioperl'};
  if(exists($redirect{chromsome})){
    $chromosome=$redirect{chromsome};
  }else{
    $chromosome=$1 if $filename=~m/chr(.*)\..*$/;
    $chromosome=$1 if $filename=~m/(\d+|.)\..*$/ and !$chromosome;
    $raw_chromosome=$chromosome;
    my %conv=(X=>-1,Y=>-2,M=>-3,Un=>-2);
    $chromosome=$conv{$chromosome} if(exists($conv{$chromosome}));
  }
  return dbFetch($redirect{ensembl_build},$raw_chromosome) if $redirect{ensembl_build};

  #extract from one of hacchy's cooked files
  print "Identified as hacchy-zome $chromosome\n";
  die "Unknown chromsome" unless ($chromosome=~m/-?\d+/ and $chromosome!=0);
  die "No annotation source file?" unless $redirect{src};
  open(my $src,$redirect{src}) or die "Couldn't open indicated src file: $redirect{src}";
  my @features;
  while(<$src>){
    chomp;
    my @a=split(/\t/);
    next unless($a[1]==$chromosome or $chromosome eq 'any');
    push(@features,new Bio::SeqFeature::Generic
	 (-primary => $out_primary,
	  -tag => {($a[0] ? (gene => $a[0]):()), ($a[0] ? (locus_tag=>$a[0]):())},
	  -start => $offset+$a[2],
	  -end => $offset+$a[3],
	  -strand => $a[4])
	);
  }
  print "Derived ".scalar(@features)." features from $redirect.\n";
  return @features;
}

sub dbFetch {
  my ($build,$chromosome)=@_;
  my $out_primary=our $src_primary;
  $out_primary=$redirect{primary} if exists $redirect{primary};
  my $offset=exists $redirect{offset} ? $redirect{offset}:0;

  use DBI;
  my $dsn = "DBI:mysql:$build:ensembldb.ensembl.org";
  my $db_user_name='anonymous';
  print "Connecting to Ensembl MySQL DB...\n";
  our $dbh = DBI->connect($dsn, $db_user_name) or die "Couldn't connect!";

  my $sth = $dbh->prepare("SELECT i.stable_id,s.name,g.seq_region_start,g.seq_region_end,g.seq_region_strand FROM gene g, gene_stable_id i, seq_region s, coord_system c WHERE c.coord_system_id=s.coord_system_id AND c.name='chromosome' AND g.gene_id=i.gene_id and s.seq_region_id=g.seq_region_id AND s.name='$chromosome'"); #weee for ugly long SQL!
  print "Connected. Running query for chromosome $chromosome...\n";
  my $rv=$sth->execute;
  print "Query returned: $rv rows\n";
  my @a;
  my @features;
  while(@a=$sth->fetchrow_array){
    push(@features,new Bio::SeqFeature::Generic
	 (-primary => $out_primary,
	  -tag => {gene => $a[0], locus_tag=>$a[0]},
	  -start => $offset+$a[2],
	  -end => $offset+$a[3],
	  -strand => $a[4])
	);
  }
  $dbh->disconnect or warn $dbh->errstr;
  return @features;
}

sub procOverrides {
  my $src=pop;
  open(my $infh,$src) or return;
  while(<$infh>){
    chomp;
    my ($key,$val)=m/^(.*?)=(.*)$/;
    print "Overriding $key as $val\n";
    $common_overrides{$key}=$val;
  }
}

sub getPath {
  my @ret=map {
    my ($name,$path,$suffix) = fileparse($_, qr{\.[^.]*});
    $path
  } @_;
  return @ret if $#_;
  return $ret[0];
}

sub slurp {
 local $/;
  open(my $fh,"@_") or return;
  return <$fh>;
}

__END__

=head1 NAME

getcds.pl - CDS extraction from various annotation formats

=head1 SYNOPSIS

cbtest.pl input file [output file] [input file [outputfile] ...]

=head1 OPTIONS

If [output file] is unspecificied, "[input file].cds" is used.
